@@ -24,7 +24,6 @@ import os
 import platform
 import stat
 import re
-import pwd
 import textwrap
 import shutil
 import subprocess
@@ -34,9 +33,9 @@ import fcntl
 import traceback
 import syslog
 import gzip
+import pwd
+
 import debconf
-import warnings
-warnings.filterwarnings("ignore", "apt API not stable yet", FutureWarning)
 import apt_pkg
 from apt.cache import Cache
 
@@ -89,37 +88,42 @@ class Install(install_misc.InstallBase):
             self.target = '/'
             return
 
-        apt_pkg.InitConfig()
-        apt_pkg.Config.set("Dir", self.target)
-        apt_pkg.Config.set("Dir::State::status",
+        apt_pkg.init_config()
+        apt_pkg.config.set("Dir", self.target)
+        apt_pkg.config.set("Dir::State::status",
                            os.path.join(self.target, 'var/lib/dpkg/status'))
-        apt_pkg.Config.set("APT::GPGV::TrustedKeyring",
+        apt_pkg.config.set("APT::GPGV::TrustedKeyring",
                            os.path.join(self.target, 'etc/apt/trusted.gpg'))
 
         # Keep this in sync with configure_apt.
         # TODO cjwatson 2011-03-03: consolidate this.
-        apt_pkg.Config.set("APT::Authentication::TrustCDROM", "true")
-        apt_pkg.Config.set("Acquire::gpgv::Options::",
+        try:
+            if self.db.get('base-installer/install-recommends') == 'false':
+                apt_pkg.config.set("APT::Install-Recommends", "false")
+        except debconf.DebconfError:
+            pass
+        apt_pkg.config.set("APT::Authentication::TrustCDROM", "true")
+        apt_pkg.config.set("Acquire::gpgv::Options::",
                            "--ignore-time-conflict")
         try:
             if self.db.get('debian-installer/allow_unauthenticated') == 'true':
-                apt_pkg.Config.set("APT::Get::AllowUnauthenticated", "true")
-                apt_pkg.Config.set(
+                apt_pkg.config.set("APT::Get::AllowUnauthenticated", "true")
+                apt_pkg.config.set(
                     "Aptitude::CmdLine::Ignore-Trust-Violations", "true")
         except debconf.DebconfError:
             pass
-        apt_pkg.Config.set("APT::CDROM::NoMount", "true")
-        apt_pkg.Config.set("Acquire::cdrom::mount", "/cdrom")
-        apt_pkg.Config.set("Acquire::cdrom::/cdrom/::Mount", "true")
-        apt_pkg.Config.set("Acquire::cdrom::/cdrom/::UMount", "true")
-        apt_pkg.Config.set("Acquire::cdrom::AutoDetect", "false")
-        apt_pkg.Config.set("Dir::Media::MountPath", "/cdrom")
+        apt_pkg.config.set("APT::CDROM::NoMount", "true")
+        apt_pkg.config.set("Acquire::cdrom::mount", "/cdrom")
+        apt_pkg.config.set("Acquire::cdrom::/cdrom/::Mount", "true")
+        apt_pkg.config.set("Acquire::cdrom::/cdrom/::UMount", "true")
+        apt_pkg.config.set("Acquire::cdrom::AutoDetect", "false")
+        apt_pkg.config.set("Dir::Media::MountPath", "/cdrom")
 
-        apt_pkg.Config.set("DPkg::Options::", "--root=%s" % self.target)
+        apt_pkg.config.set("DPkg::Options::", "--root=%s" % self.target)
         # We don't want apt-listchanges or dpkg-preconfigure, so just clear
         # out the list of pre-installation hooks.
-        apt_pkg.Config.clear("DPkg::Pre-Install-Pkgs")
-        apt_pkg.InitSystem()
+        apt_pkg.config.clear("DPkg::Pre-Install-Pkgs")
+        apt_pkg.init_system()
 
         use_restricted = True
         try:
@@ -148,6 +152,8 @@ class Install(install_misc.InstallBase):
 
         self.db.progress('START', self.start, self.end, 'ubiquity/install/title')
 
+        self.configure_python()
+
         self.next_region()
         self.db.progress('INFO', 'ubiquity/install/network')
         self.configure_network()
@@ -157,6 +163,7 @@ class Install(install_misc.InstallBase):
         self.configure_apt()
 
         self.configure_plugins()
+        self.configure_face()
 
         self.next_region()
         self.run_target_config_hooks()
@@ -230,6 +237,15 @@ class Install(install_misc.InstallBase):
                 syslog.syslog(syslog.LOG_WARNING, line)
             self.db.input('critical', 'ubiquity/install/broken_network_copy')
             self.db.go()
+        try:
+            self.copy_bluetooth_config()
+        except:
+            syslog.syslog(syslog.LOG_WARNING,
+                'Could not copy the bluetooth configuration:')
+            for line in traceback.format_exc().split('\n'):
+                syslog.syslog(syslog.LOG_WARNING, line)
+            self.db.input('critical', 'ubiquity/install/broken_bluetooth_copy')
+            self.db.go()
         try:
             self.recache_apparmor()
         except:
@@ -252,6 +268,118 @@ class Install(install_misc.InstallBase):
 
         self.db.progress('SET', self.end)
 
+    def configure_face(self):
+        PHOTO_PATH = '/var/lib/ubiquity/webcam_photo.png'
+        target_user = self.db.get('passwd/username')
+        uid = subprocess.Popen(['chroot', self.target, 'sudo', '-u',
+            target_user, '--', 'id', '-u'], stdout=subprocess.PIPE)
+        uid = uid.communicate()[0].strip('\n')
+        gid = subprocess.Popen(['chroot', self.target, 'sudo', '-u',
+            target_user, '--', 'id', '-g'], stdout=subprocess.PIPE)
+        gid = gid.communicate()[0].strip('\n')
+        try:
+            uid = int(uid)
+            gid = int(gid)
+        except ValueError:
+            return
+        if os.path.exists(PHOTO_PATH):
+            targetpath = os.path.join(self.target, 'home', target_user, '.face')
+            shutil.copy2(PHOTO_PATH, targetpath)
+            os.lchown(targetpath, uid, gid)
+
+    def configure_python(self):
+        """Byte-compile Python modules.
+
+        To save space, Ubuntu excludes .pyc files from the live filesystem.
+        Recreate them now to restore the appearance of a system installed
+        from .debs."""
+
+        cache = Cache()
+
+        # Python standard library.
+        re_minimal = re.compile('^python\d+\.\d+-minimal$')
+        python_installed = sorted(
+            [pkg[:-8] for pkg in cache.keys()
+                      if re_minimal.match(pkg) and cache[pkg].is_installed])
+        for python in python_installed:
+            re_file = re.compile('^/usr/lib/%s/.*\.py$' % python)
+            files = [f for f in cache['%s-minimal' % python].installed_files
+                       if re_file.match(f) and
+                          not os.path.exists(os.path.join(self.target,
+                                                          '%sc' % f[1:]))]
+            install_misc.chrex(self.target, python,
+                               '/usr/lib/%s/py_compile.py' % python, *files)
+            files = [f for f in cache[python].installed_files
+                       if re_file.match(f) and
+                          not os.path.exists(os.path.join(self.target,
+                                                          '%sc' % f[1:]))]
+            install_misc.chrex(self.target, python,
+                               '/usr/lib/%s/py_compile.py' % python, *files)
+
+        # Modules provided by the core Debian Python packages.
+        default = subprocess.Popen(
+            ['chroot', self.target, 'pyversions', '-d'],
+            stdout=subprocess.PIPE).communicate()[0].rstrip('\n')
+        if default:
+            install_misc.chrex(self.target, default, '-m', 'compileall',
+                               '/usr/share/python/')
+        if osextras.find_on_path_root(self.target, 'py3compile'):
+            install_misc.chrex(self.target, 'py3compile', '-p', 'python3',
+                               '/usr/share/python3/')
+
+        def run_hooks(path, *args):
+            for hook in osextras.glob_root(self.target, path):
+                if not os.access(os.path.join(self.target, hook[1:]), os.X_OK):
+                    continue
+                install_misc.chrex(self.target, hook, *args)
+
+        # Public and private modules provided by other packages.
+        install_misc.chroot_setup(self.target)
+        try:
+            if osextras.find_on_path_root(self.target, 'pyversions'):
+                supported = subprocess.Popen(
+                    ['chroot', self.target, 'pyversions', '-s'],
+                    stdout=subprocess.PIPE).communicate()[0].rstrip('\n')
+                for python in supported.split():
+                    try:
+                        cachedpython = cache['%s-minimal' % python]
+                    except KeyError:
+                        continue
+                    if not cachedpython.is_installed:
+                        continue
+                    version = cachedpython.installed.version
+                    run_hooks('/usr/share/python/runtime.d/*.rtinstall',
+                              'rtinstall', python, '', version)
+                    run_hooks('/usr/share/python/runtime.d/*.rtupdate',
+                              'pre-rtupdate', python, python)
+                    run_hooks('/usr/share/python/runtime.d/*.rtupdate',
+                              'rtupdate', python, python)
+                    run_hooks('/usr/share/python/runtime.d/*.rtupdate',
+                              'post-rtupdate', python, python)
+
+            if osextras.find_on_path_root(self.target, 'py3versions'):
+                supported = subprocess.Popen(
+                    ['chroot', self.target, 'py3versions', '-s'],
+                    stdout=subprocess.PIPE).communicate()[0].rstrip('\n')
+                for python in supported.split():
+                    try:
+                        cachedpython = cache['%s-minimal' % python]
+                    except KeyError:
+                        continue
+                    if not cachedpython.is_installed:
+                        continue
+                    version = cachedpython.installed.version
+                    run_hooks('/usr/share/python3/runtime.d/*.rtinstall',
+                              'rtinstall', python, '', version)
+                    run_hooks('/usr/share/python3/runtime.d/*.rtupdate',
+                              'pre-rtupdate', python, python)
+                    run_hooks('/usr/share/python3/runtime.d/*.rtupdate',
+                              'rtupdate', python, python)
+                    run_hooks('/usr/share/python3/runtime.d/*.rtupdate',
+                              'post-rtupdate', python, python)
+        finally:
+            install_misc.chroot_cleanup(self.target)
+
     def configure_network(self):
         """Automatically configure the network.
 
@@ -270,7 +398,27 @@ class Install(install_misc.InstallBase):
         if self.target != '/':
             for path in ('/etc/network/interfaces', '/etc/resolv.conf'):
                 if os.path.exists(path):
-                    shutil.copy2(path, os.path.join(self.target, path[1:]))
+                    targetpath = os.path.join(self.target, path[1:])
+                    st = os.lstat(path)
+                    if stat.S_ISLNK(st.st_mode):
+                        if os.path.lexists(targetpath):
+                            os.unlink(targetpath)
+                        linkto = os.readlink(path)
+                        os.symlink(linkto, targetpath)
+                    else:
+                        shutil.copy2(path, os.path.join(self.target, path[1:]))
+        else:
+            if not os.path.exists('/etc/network/interfaces'):
+                # Make sure there's at least something here so that ifupdown
+                # doesn't get upset at boot.
+                with open('/etc/network/interfaces', 'w') as interfaces:
+                    print >>interfaces, textwrap.dedent("""\
+                        # This file describes the network interfaces available on your system
+                        # and how to activate them. For more information, see interfaces(5).
+
+                        # The loopback network interface
+                        auto lo
+                        iface lo inet loopback""")
 
         try:
             hostname = self.db.get('netcfg/get_hostname')
@@ -408,6 +556,17 @@ class Install(install_misc.InstallBase):
 
         # Keep this in sync with __init__.
 
+        try:
+            if self.db.get('base-installer/install-recommends') == 'false':
+                apt_conf_ir = open(
+                    os.path.join(self.target,
+                                 'etc/apt/apt.conf.d/00InstallRecommends'),
+                    'w')
+                print >>apt_conf_ir, 'APT::Install-Recommends "false";'
+                apt_conf_ir.close()
+        except debconf.DebconfError:
+            pass
+
         # Make apt trust CDs. This is not on by default (we think).
         # This will be left in place on the installed system.
         apt_conf_tc = open(os.path.join(
@@ -494,6 +653,7 @@ class Install(install_misc.InstallBase):
             self.db.progress('INFO', 'ubiquity/install/target_hooks')
             for hookentry in hooks:
                 hook = os.path.join(hookdir, hookentry)
+                syslog.syslog('running %s' % hook)
                 if not os.access(hook, os.X_OK):
                     self.db.progress('STEP', 1)
                     continue
@@ -550,7 +710,7 @@ class Install(install_misc.InstallBase):
         kern = install_misc.get_cache_pkg(cache, pkg)
         if kern is None:
             return None
-        pkc = cache._depcache.GetCandidateVer(kern._pkg)
+        pkc = cache._depcache.get_candidate_ver(kern._pkg)
         if 'Depends' in pkc.depends_list:
             dependencies = pkc.depends_list['Depends']
         else:
@@ -855,7 +1015,7 @@ class Install(install_misc.InstallBase):
                                 self.db.set('grub-installer/bootdev', response)
                         else:
                             break
-                elif (arch == 'armel' and
+                elif (arch in ('armel', 'armhf') and
                       subarch in ('dove', 'imx51', 'iop32x', 'ixp4xx', 'orion5x', 'omap')):
                     from ubiquity.components import flash_kernel
                     dbfilter = flash_kernel.FlashKernel(None, self.db)
@@ -961,7 +1121,8 @@ class Install(install_misc.InstallBase):
         except debconf.DebconfError:
             inst_langpacks = False
         if inst_langpacks:
-            self.langpacks = self.select_language_packs()
+            self.select_language_packs()
+            recorded = install_misc.query_recorded_installed()
 
         try:
             extra_packages = self.db.get('oem-config/extra_packages')
@@ -976,7 +1137,7 @@ class Install(install_misc.InstallBase):
                 return
 
         if inst_langpacks:
-            extra_packages += self.langpacks
+            extra_packages += recorded
 
         save_replace = None
         save_override = None
@@ -1042,8 +1203,8 @@ class Install(install_misc.InstallBase):
     def install_restricted_extras(self):
         if self.db.get('ubiquity/use_nonfree') == 'true':
             self.db.progress('INFO', 'ubiquity/install/nonfree')
-            package = self.db.get('ubiquity/nonfree_package')
-            self.do_install([package])
+            packages = self.db.get('ubiquity/nonfree_package').split()
+            self.do_install(packages)
             try:
                 install_misc.chrex(self.target,'dpkg-divert', '--package',
                         'ubiquity', '--rename', '--quiet', '--add',
@@ -1191,10 +1352,13 @@ class Install(install_misc.InstallBase):
                 sourcepath = os.path.join(source, relpath)
                 targetpath = os.path.join(target, relpath)
                 st = os.lstat(sourcepath)
+
+                # Remove the target if necessary and if we can.
+                install_misc.remove_target(source, target, relpath, st)
+
+                # Now actually copy source to target.
                 mode = stat.S_IMODE(st.st_mode)
                 if stat.S_ISLNK(st.st_mode):
-                    if os.path.lexists(targetpath):
-                        os.unlink(targetpath)
                     linkto = os.readlink(sourcepath)
                     os.symlink(linkto, targetpath)
                 elif stat.S_ISDIR(st.st_mode):
@@ -1209,7 +1373,6 @@ class Install(install_misc.InstallBase):
                 elif stat.S_ISSOCK(st.st_mode):
                     os.mknod(targetpath, stat.S_IFSOCK | mode)
                 elif stat.S_ISREG(st.st_mode):
-                    osextras.unlink_force(targetpath)
                     install_misc.copy_file(self.db, sourcepath, targetpath, True)
 
                 os.lchown(targetpath, uid, gid)
@@ -1219,7 +1382,11 @@ class Install(install_misc.InstallBase):
                     directory_times.append((targetpath, st.st_atime, st.st_mtime))
                 # os.utime() sets timestamp of target, not link
                 elif not stat.S_ISLNK(st.st_mode):
-                    os.utime(targetpath, (st.st_atime, st.st_mtime))
+                    try:
+                        os.utime(targetpath, (st.st_atime, st.st_mtime))
+                    except Exception:
+                        # We can live with timestamps being wrong.
+                        pass
 
         # Apply timestamps to all directories now that the items within them
         # have been copied.
@@ -1227,7 +1394,7 @@ class Install(install_misc.InstallBase):
             (directory, atime, mtime) = dirtime
             try:
                 os.utime(directory, (atime, mtime))
-            except OSError:
+            except Exception:
                 # I have no idea why I've been getting lots of bug reports
                 # about this failing, but I really don't care. Ignore it.
                 pass
@@ -1239,12 +1406,30 @@ class Install(install_misc.InstallBase):
         # Looking through files for packages to remove is pretty quick, so
         # don't bother with a progress bar for that.
 
-        # Check for packages specific to the live CD.
+        # Check for packages specific to the live CD.  (manifest-desktop is
+        # the old method, which listed all the packages to keep;
+        # manifest-remove is the new method, which lists all the packages to
+        # remove.)
+        manifest_remove = os.path.join(self.casper_path,
+                                       'filesystem.manifest-remove')
         manifest_desktop = os.path.join(self.casper_path,
                                         'filesystem.manifest-desktop')
         manifest = os.path.join(self.casper_path, 'filesystem.manifest')
-        if (os.path.exists(manifest_desktop) and
-            os.path.exists(manifest)):
+        if os.path.exists(manifest_remove) and os.path.exists(manifest):
+            difference = set()
+            manifest_file = open(manifest_remove)
+            for line in manifest_file:
+                if line.strip() != '' and not line.startswith('#'):
+                    difference.add(line.split()[0])
+            manifest_file.close()
+            live_packages = set()
+            manifest_file = open(manifest)
+            for line in manifest_file:
+                if line.strip() != '' and not line.startswith('#'):
+                    live_packages.add(line.split()[0])
+            manifest_file.close()
+            desktop_packages = live_packages - difference
+        elif os.path.exists(manifest_desktop) and os.path.exists(manifest):
             desktop_packages = set()
             manifest_file = open(manifest_desktop)
             for line in manifest_file:
@@ -1311,7 +1496,9 @@ class Install(install_misc.InstallBase):
 
         if oem_remove_extras:
             installed = (desktop_packages | keep - regular - recursive)
-            p = os.path.join(self.target, '/var/lib/ubiquity/installed-packages')
+            if not os.path.exists(os.path.join(self.target, 'var/lib/ubiquity')):
+                os.makedirs(os.path.join(self.target, 'var/lib/ubiquity'))
+            p = os.path.join(self.target, 'var/lib/ubiquity/installed-packages')
             with open(p, 'w') as fp:
                 for line in installed:
                     print >>fp, line
@@ -1343,56 +1530,50 @@ class Install(install_misc.InstallBase):
     def copy_network_config(self):
         if 'UBIQUITY_OEM_USER_CONFIG' in os.environ:
             return
+        try:
+            if self.db.get('oem-config/enable') == 'true':
+                return
+        except debconf.DebconfError:
+            pass
 
-        if 'SUDO_USER' in os.environ:
-            casper_user = os.path.expanduser('~%s' % os.environ['SUDO_USER'])
-        else:
-            casper_user = os.path.expanduser('~')
-        target_user = self.db.get('passwd/username')
-
-        # GTK
-        homedir = '/home/%s' % target_user
-        target_keyrings = os.path.join(self.target, homedir[1:],
-                                       '.gnome2/keyrings')
+        source_nm = "/etc/NetworkManager/system-connections/"
+        target_nm = "/target/etc/NetworkManager/system-connections/"
 
         # Sanity checks.  We don't want to do anything if a network
-        # configuration already exists, which will be the case if the user
-        # selected to install without formatting.
-        if os.path.exists(target_keyrings):
-            return
-        config_source = 'xml:readwrite:%s/.gconf' % homedir
-        subp = subprocess.Popen(['chroot', self.target, 'sudo', '-i', '-n',
-            '-u', target_user, '--', 'gconftool-2', '--direct',
-            '--config-source', config_source, '--dir-exists',
-            '/system/networking'], close_fds=True)
-        subp.communicate()
-        if subp.returncode == 0:
-            return
+        # configuration already exists on the target
+        if os.path.exists(source_nm) and os.path.exists(target_nm):
+            for network in os.listdir(source_nm):
+                # Skip LTSP live
+                if network == "LTSP":
+                    continue
 
-        from ubiquity import gconftool
-        if gconftool.dump('/system/networking', os.path.join(self.target,
-                          'tmp/live-network-config')):
-            # Ick.
-            subprocess.call(['log-output', '-t', 'ubiquity', 'chroot',
-                self.target, 'sudo', '-i', '-n', '-u', target_user, '--',
-                'gconftool-2', '--direct', '--config-source', config_source,
-                '--load', '/tmp/live-network-config'], close_fds=True)
-            os.remove('/target/tmp/live-network-config')
-            source_keyrings = os.path.join(casper_user, '.gnome2/keyrings')
-            if os.path.exists(source_keyrings):
-                # We could just figure out what $HOME is and stat it as an
-                # alternative.
-                uid = subprocess.Popen(['chroot', self.target, 'sudo', '-u',
-                    target_user, '--', 'id', '-u'],
-                    stdout=subprocess.PIPE).communicate()[0].strip('\n')
-                gid = subprocess.Popen(['chroot', self.target, 'sudo', '-u',
-                    target_user, '--', 'id', '-g'],
-                    stdout=subprocess.PIPE).communicate()[0].strip('\n')
-                uid = int(uid)
-                gid = int(gid)
-                self.copy_tree(source_keyrings, target_keyrings, uid, gid)
+                source_network = os.path.join(source_nm, network)
+                target_network = os.path.join(target_nm, network)
 
-        # KDE TODO
+                if os.path.exists(target_network):
+                    continue
+
+                shutil.copy(source_network, target_network)
+
+    def copy_bluetooth_config(self):
+        if 'UBIQUITY_OEM_USER_CONFIG' in os.environ:
+            return
+        try:
+            if self.db.get('oem-config/enable') == 'true':
+                return
+        except debconf.DebconfError:
+            pass
+
+        source_bluetooth = "/var/lib/bluetooth/"
+        target_bluetooth = "/target/var/lib/bluetooth/"
+
+        # Ensure the target doesn't exist
+        if os.path.exists(target_bluetooth):
+            shutil.rmtree(target_bluetooth)
+
+        # Copy /var/lib/bluetooth to /target/var/lib/bluetooth/
+        if os.path.exists(source_bluetooth):
+            shutil.copytree(source_bluetooth, target_bluetooth)
 
     def recache_apparmor(self):
         """Generate an apparmor cache in /etc/apparmor.d/cache to speed up boot
@@ -1421,7 +1602,13 @@ class Install(install_misc.InstallBase):
 
         # we don't use copy_network_config casper user trick as it's not and not
         # ubuntu in install mode
-        casper_user = 'ubuntu'
+        try:
+            casper_user = pwd.getpwuid(999).pw_name
+        except KeyError:
+            # We're on a weird system where the casper user isn't uid 999
+            # just stop there
+            return
+
         casper_user_home = os.path.expanduser('~%s' % casper_user)
         casper_user_wallpaper_cache_dir = os.path.join(casper_user_home,
                                                        '.cache', 'wallpaper')
@@ -1539,6 +1726,7 @@ class Install(install_misc.InstallBase):
                 self.target, 'etc/apt/apt.conf.d', apt_conf))
 
 if __name__ == '__main__':
+    os.environ['DPKG_UNTRANSLATED_MESSAGES'] = '1'
     if not os.path.exists('/var/lib/ubiquity'):
         os.makedirs('/var/lib/ubiquity')
 
